from .download import download_zip
from pathlib import Path
down_path = Path(__file__).parent
xz_url = "addons_file"+"/"+down_path.name+"/"
download_zip(xz_url + "TerraNode_library.zip",down_path)


import bpy
import os
import struct
from bpy.types import Scene
from bpy.props import StringProperty, IntProperty, EnumProperty, CollectionProperty, PointerProperty, BoolProperty
from bpy.app.handlers import persistent
from math import sqrt
from mathutils import Vector, Matrix
from pathlib import Path
import numpy as np

bl_info = {
    "name": "DJH TerraNode",
    "description": "Helpful tools to extend the TerraNode toolkit.",
    "author": "xz-blender repack",
    "version": (1, 3, 0),
    "blender": (4, 1, 0),
    "category": "3D View",
    "doc_url": "https://danielhickox.com/terranode#exporter",
}

classes = []
class DJH_OT_RenameTerrains(bpy.types.Operator):
    """Saves the selected TerraNode terrain as a set of texture maps."""
    bl_idname = 'djh.terranode_auto_name'
    bl_label = 'Auto-Name Terrains'
    bl_options = {'REGISTER', 'UNDO'}

    prefix: StringProperty(name="Prefix", default="terrain")
    
    def execute(self, context):        
        def vec_min(a: Vector, b: Vector):
            return Vector((min(a.x,b.x), min(a.y,b.y), min(a.z, b.z)))
        def vec_max(a: Vector, b: Vector):
            return Vector((max(a.x,b.x), max(a.y,b.y), max(a.z, b.z)))
        def vec_div(a: Vector, b: Vector):
            vec = Vector((0,0,0))
            if b.x != 0: vec.x = a.x / b.x
            if b.y != 0: vec.y = a.y / b.y
            if b.z != 0: vec.z = a.z / b.z
            return vec

        collection = context.object.users_collection[0]

        terrains = []
        loc_min = context.object.location
        loc_max = context.object.location

        len_x = []
        len_y = []

        for obj in collection.objects:
            base = None
            for m in obj.modifiers:
                if m.type == 'NODES':
                    if m.node_group.name == '  ⛰ Base':
                        base = m
                        break
            
            if base == None:
                print('--not a terrain, skipping')
                continue

            terrains.append(obj)
            loc_min = vec_min(loc_min, obj.location)
            loc_max = vec_max(loc_max, obj.location)

            if obj.location.x not in len_x:
                len_x.append(obj.location.x)

            if obj.location.y not in len_y:
                len_y.append(obj.location.y)

        size = loc_max - loc_min
        scale = Vector((len(len_x)-1, len(len_y)-1, 0))
        for i in range(len(terrains)):
            co = vec_div(terrains[i].location - loc_min, size) * scale

            if bpy.context.preferences.addons[__package__].preferences.export_target == 'unreal':
                terrains[i].name = f"{self.prefix}_x{int(co.x)}_y{int(scale.y - co.y)}"
            else:
                terrains[i].name = f"{self.prefix}_{int(co.x)}_{int(co.y)}"

        return {'FINISHED'}

classes.append(DJH_OT_RenameTerrains)


class DJH_OT_ExportTerraNode(bpy.types.Operator):
    """Saves the selected TerraNode terrain as a set of texture maps."""
    bl_idname = 'djh.terranode_export'
    bl_label = 'Export TerraNode'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        
        def GetCleanupNode():
            if not '|DJH_⛰_BakePrep' in bpy.data.node_groups:
                if bpy.app.version[0] == 4 and bpy.app.version[1] > 0:
                    group = bpy.data.node_groups.new('|DJH_⛰_BakePrep', 'GeometryNodeTree')
                    group.is_modifier = True

                    group.interface.new_socket(socket_type='NodeSocketGeometry', name='Geometry', in_out='OUTPUT')
                    group.interface.new_socket(socket_type='NodeSocketGeometry', name='Geometry', in_out='INPUT')
                    group.interface.new_socket(socket_type='NodeSocketCollection', name='Collection', in_out='INPUT')

                    n0 = group.nodes.new('GeometryNodeInputNamedAttribute')
                    n0.name = 'Named Attribute.001'
                    n0.location = Vector((-537.2661743164062,-321.472412109375))
                    n0.data_type = 'INT'
                    n0.inputs[0].default_value = 'terrain_type'

                    n0 = group.nodes.new('GeometryNodeCollectionInfo')
                    n0.name = 'Collection Info'
                    n0.location = Vector((-749.5956420898438,-223.84791564941406))

                    n0 = group.nodes.new('GeometryNodeRealizeInstances')
                    n0.name = 'Realize Instances'
                    n0.location = Vector((-541.3276977539062,-232.00579833984375))

                    n0 = group.nodes.new('GeometryNodeDeleteGeometry')
                    n0.name = 'Delete Geometry.001'
                    n0.location = Vector((-357.2408142089844,-229.15664672851562))
                    n0.domain = 'FACE'
                    n0.mode = 'ALL'

                    n0 = group.nodes.new('GeometryNodeMergeByDistance')
                    n0.name = 'Merge by Distance'
                    n0.location = Vector((263.8589782714844,-211.6722412109375))
                    n0.mode = 'ALL'
                    n0.inputs[2].default_value = 0.0010000000474974513

                    n0 = group.nodes.new('ShaderNodeVectorMath')
                    n0.name = 'Vector Math'
                    n0.location = Vector((-147.089599609375,-415.6555480957031))
                    n0.operation = 'MULTIPLY'
                    n0.inputs[1].default_value =  Vector((1.0,1.0,0.0))

                    n0 = group.nodes.new('GeometryNodeSetPosition')
                    n0.name = 'Set Position'
                    n0.location = Vector((56.693260192871094,-200.30648803710938))
                    n0.inputs[3].default_value =  Vector((0.0,0.0,0.0))

                    n0 = group.nodes.new('GeometryNodeSampleNearest')
                    n0.name = 'Sample Nearest'
                    n0.location = Vector((449.42498779296875,-351.94842529296875))
                    n0.domain = 'POINT'
                    n0.inputs[1].default_value =  Vector((0.0,0.0,0.0))

                    n0 = group.nodes.new('GeometryNodeCaptureAttribute')
                    n0.name = 'Capture Attribute'
                    n0.location = Vector((-148.33592224121094,-225.3878173828125))
                    n0.data_type = 'FLOAT'
                    n0.domain = 'POINT'

                    n0 = group.nodes.new('GeometryNodeInputNamedAttribute')
                    n0.name = 'Named Attribute.002'
                    n0.location = Vector((-350.85650634765625,-392.1100769042969))
                    n0.data_type = 'FLOAT'
                    n0.inputs[0].default_value = 'elevation'

                    n0 = group.nodes.new('GeometryNodeInputPosition')
                    n0.name = 'Position'
                    n0.location = Vector((-353.49560546875,-531.7010498046875))

                    n0 = group.nodes.new('GeometryNodeSampleIndex')
                    n0.name = 'Sample Index'
                    n0.location = Vector((452.49163818359375,-141.03515625))
                    n0.data_type = 'FLOAT'
                    n0.domain = 'POINT'

                    n0 = group.nodes.new('GeometryNodeStoreNamedAttribute')
                    n0.name = 'Store Named Attribute'
                    n0.location = Vector((715.5792846679688,39.32191467285156))
                    n0.data_type = 'FLOAT'
                    n0.domain = 'POINT'
                    n0.inputs[2].default_value = 'elevation'

                    n0 = group.nodes.new('NodeGroupOutput')
                    n0.name = 'Group Output'
                    n0.location = Vector((1552.889892578125,-9.163956642150879))

                    n0 = group.nodes.new('GeometryNodeDeleteGeometry')
                    n0.name = 'Delete Geometry'
                    n0.location = Vector((450.5357666015625,37.465576171875))
                    n0.domain = 'FACE'
                    n0.mode = 'ALL'

                    n0 = group.nodes.new('NodeGroupInput')
                    n0.name = 'Group Input'
                    n0.location = Vector((265.4375,48.17465591430664))

                    n0 = group.nodes.new('GeometryNodeInputNamedAttribute')
                    n0.name = 'Named Attribute'
                    n0.location = Vector((268.4526672363281,-62.321258544921875))
                    n0.data_type = 'INT'
                    n0.inputs[0].default_value = 'terrain_type'

                    n0 = group.nodes.new('NodeGroupInput')
                    n0.name = 'Group Input.001'
                    n0.location = Vector((-960.2911987304688,-271.26739501953125))

                    n0 = group.nodes.new('GeometryNodeSortElements')
                    n0.name = 'Sort Elements'
                    n0.location = Vector((1131.029541015625,48.87976837158203))
                    n0.domain = 'POINT'
                    n0.inputs[2].default_value = 0

                    n0 = group.nodes.new('GeometryNodeInputPosition')
                    n0.name = 'Position.001'
                    n0.location = Vector((729.5863037109375,-166.6861572265625))

                    n0 = group.nodes.new('ShaderNodeSeparateXYZ')
                    n0.name = 'Separate XYZ'
                    n0.location = Vector((923.701171875,-101.22299194335938))

                    n0 = group.nodes.new('GeometryNodeSortElements')
                    n0.name = 'Sort Elements.001'
                    n0.location = Vector((1340.0367431640625,-7.137977600097656))
                    n0.domain = 'POINT'
                    n0.inputs[2].default_value = 0



                    group.links.new(group.nodes['Group Input'].outputs[0], group.nodes['Delete Geometry'].inputs[0])
                    group.links.new(group.nodes['Collection Info'].outputs[0], group.nodes['Realize Instances'].inputs[0])
                    group.links.new(group.nodes['Realize Instances'].outputs[0], group.nodes['Delete Geometry.001'].inputs[0])
                    group.links.new(group.nodes['Delete Geometry.001'].outputs[0], group.nodes['Capture Attribute'].inputs[0])
                    group.links.new(group.nodes['Capture Attribute'].outputs[0], group.nodes['Set Position'].inputs[0])
                    group.links.new(group.nodes['Position'].outputs[0], group.nodes['Vector Math'].inputs[0])
                    group.links.new(group.nodes['Vector Math'].outputs[0], group.nodes['Set Position'].inputs[2])
                    group.links.new(group.nodes['Set Position'].outputs[0], group.nodes['Merge by Distance'].inputs[0])
                    group.links.new(group.nodes['Merge by Distance'].outputs[0], group.nodes['Sample Nearest'].inputs[0])
                    group.links.new(group.nodes['Merge by Distance'].outputs[0], group.nodes['Sample Index'].inputs[0])
                    group.links.new(group.nodes['Delete Geometry'].outputs[0], group.nodes['Store Named Attribute'].inputs[0])
                    group.links.new(group.nodes['Capture Attribute'].outputs[1], group.nodes['Sample Index'].inputs[1])
                    group.links.new(group.nodes['Named Attribute.002'].outputs[0], group.nodes['Capture Attribute'].inputs[1])
                    group.links.new(group.nodes['Sample Index'].outputs[0], group.nodes['Store Named Attribute'].inputs[3])
                    group.links.new(group.nodes['Group Input.001'].outputs[1], group.nodes['Collection Info'].inputs[0])
                    group.links.new(group.nodes['Sample Nearest'].outputs[0], group.nodes['Sample Index'].inputs[2])
                    group.links.new(group.nodes['Named Attribute.001'].outputs[0], group.nodes['Delete Geometry.001'].inputs[1])
                    group.links.new(group.nodes['Named Attribute'].outputs[0], group.nodes['Delete Geometry'].inputs[1])
                    group.links.new(group.nodes['Store Named Attribute'].outputs[0], group.nodes['Sort Elements'].inputs[0])
                    group.links.new(group.nodes['Sort Elements.001'].outputs[0], group.nodes['Group Output'].inputs[0])
                    group.links.new(group.nodes['Position.001'].outputs[0], group.nodes['Separate XYZ'].inputs[0])
                    group.links.new(group.nodes['Separate XYZ'].outputs[0], group.nodes['Sort Elements'].inputs[3])
                    group.links.new(group.nodes['Sort Elements'].outputs[0], group.nodes['Sort Elements.001'].inputs[0])
                    group.links.new(group.nodes['Separate XYZ'].outputs[1], group.nodes['Sort Elements.001'].inputs[3])

                else:

                    tree = bpy.data.node_groups.new('|DJH_⛰_BakePrep', 'GeometryNodeTree')
                    tree.is_modifier = True

                    tree.interface.new_socket(socket_type='NodeSocketGeometry', name='Geometry', in_out='OUTPUT')
                    tree.interface.new_socket(socket_type='NodeSocketGeometry', name='Geometry', in_out='INPUT')
                    tree.interface.new_socket(socket_type='NodeSocketCollection', name='Collection', in_out='INPUT')

                    n = tree.nodes.new('GeometryNodeInputNamedAttribute')
                    n.name = 'Named Attribute.001'
                    n.location = Vector((-556.1599731445312,-302.54925537109375))
                    n.data_type = 'INT'
                    n.inputs[0].default_value = 'terrain_type'

                    n = tree.nodes.new('GeometryNodeCollectionInfo')
                    n.name = 'Collection Info'
                    n.location = Vector((-768.4894409179688,-204.9247589111328))
                    n.transform_space = 'RELATIVE'

                    n = tree.nodes.new('GeometryNodeRealizeInstances')
                    n.name = 'Realize Instances'
                    n.location = Vector((-560.2214965820312,-213.0826416015625))

                    n = tree.nodes.new('GeometryNodeDeleteGeometry')
                    n.name = 'Delete Geometry.001'
                    n.location = Vector((-376.1346130371094,-210.23348999023438))
                    n.domain = 'FACE'
                    n.mode = 'ALL'

                    n = tree.nodes.new('GeometryNodeMergeByDistance')
                    n.name = 'Merge by Distance'
                    n.location = Vector((244.96517944335938,-192.74908447265625))
                    n.mode = 'ALL'
                    n.inputs[2].default_value = 0.0010000000474974513

                    n = tree.nodes.new('ShaderNodeVectorMath')
                    n.name = 'Vector Math'
                    n.location = Vector((-165.98341369628906,-396.7323913574219))
                    n.operation = 'MULTIPLY'
                    n.inputs[1].default_value =  Vector((1.0,1.0,0.0))

                    n = tree.nodes.new('GeometryNodeSetPosition')
                    n.name = 'Set Position'
                    n.location = Vector((37.7994499206543,-181.38333129882812))
                    n.inputs[3].default_value =  Vector((0.0,0.0,0.0))

                    n = tree.nodes.new('GeometryNodeSampleNearest')
                    n.name = 'Sample Nearest'
                    n.location = Vector((429.45208740234375,-331.9455871582031))
                    n.domain = 'POINT'
                    n.inputs[1].default_value =  Vector((0.0,0.0,0.0))

                    n = tree.nodes.new('GeometryNodeCaptureAttribute')
                    n.name = 'Capture Attribute'
                    n.location = Vector((-167.229736328125,-206.46466064453125))
                    n.data_type = 'FLOAT'
                    n.domain = 'POINT'

                    n = tree.nodes.new('GeometryNodeInputNamedAttribute')
                    n.name = 'Named Attribute.002'
                    n.location = Vector((-369.75030517578125,-373.1869201660156))
                    n.data_type = 'FLOAT'
                    n.inputs[0].default_value = 'elevation'

                    n = tree.nodes.new('GeometryNodeInputPosition')
                    n.name = 'Position'
                    n.location = Vector((-372.389404296875,-512.7778930664062))

                    n = tree.nodes.new('GeometryNodeSampleIndex')
                    n.name = 'Sample Index'
                    n.location = Vector((433.59783935546875,-122.11200714111328))
                    n.data_type = 'FLOAT'
                    n.domain = 'POINT'

                    n = tree.nodes.new('GeometryNodeStoreNamedAttribute')
                    n.name = 'Store Named Attribute'
                    n.location = Vector((696.6854858398438,58.24506378173828))
                    n.data_type = 'FLOAT'
                    n.domain = 'POINT'
                    n.inputs[2].default_value = 'elevation'

                    n = tree.nodes.new('NodeGroupOutput')
                    n.name = 'Group Output'
                    n.location = Vector((987.7073974609375,50.81420135498047))

                    n = tree.nodes.new('GeometryNodeDeleteGeometry')
                    n.name = 'Delete Geometry'
                    n.location = Vector((431.6419677734375,56.38872528076172))
                    n.domain = 'FACE'
                    n.mode = 'ALL'

                    n = tree.nodes.new('NodeGroupInput')
                    n.name = 'Group Input'
                    n.location = Vector((225.760498046875,76.55937957763672))

                    n = tree.nodes.new('GeometryNodeInputNamedAttribute')
                    n.name = 'Named Attribute'
                    n.location = Vector((222.162841796875,-26.36726951599121))
                    n.data_type = 'INT'
                    n.inputs[0].default_value = 'terrain_type'

                    n = tree.nodes.new('NodeGroupInput')
                    n.name = 'Group Input.001'
                    n.location = Vector((-979.1849975585938,-252.34425354003906))

                    tree.links.new(tree.nodes['Group Input'].outputs[0], tree.nodes['Delete Geometry'].inputs[0])
                    tree.links.new(tree.nodes['Named Attribute'].outputs[4], tree.nodes['Delete Geometry'].inputs[1])
                    tree.links.new(tree.nodes['Collection Info'].outputs[0], tree.nodes['Realize Instances'].inputs[0])
                    tree.links.new(tree.nodes['Named Attribute.001'].outputs[4], tree.nodes['Delete Geometry.001'].inputs[1])
                    tree.links.new(tree.nodes['Realize Instances'].outputs[0], tree.nodes['Delete Geometry.001'].inputs[0])
                    tree.links.new(tree.nodes['Delete Geometry.001'].outputs[0], tree.nodes['Capture Attribute'].inputs[0])
                    tree.links.new(tree.nodes['Position'].outputs[0], tree.nodes['Capture Attribute'].inputs[1])
                    tree.links.new(tree.nodes['Capture Attribute'].outputs[0], tree.nodes['Set Position'].inputs[0])
                    tree.links.new(tree.nodes['Position'].outputs[0], tree.nodes['Vector Math'].inputs[0])
                    tree.links.new(tree.nodes['Vector Math'].outputs[0], tree.nodes['Set Position'].inputs[2])
                    tree.links.new(tree.nodes['Set Position'].outputs[0], tree.nodes['Merge by Distance'].inputs[0])
                    tree.links.new(tree.nodes['Merge by Distance'].outputs[0], tree.nodes['Sample Nearest'].inputs[0])
                    tree.links.new(tree.nodes['Sample Nearest'].outputs[0], tree.nodes['Sample Index'].inputs[7])
                    tree.links.new(tree.nodes['Merge by Distance'].outputs[0], tree.nodes['Sample Index'].inputs[0])
                    tree.links.new(tree.nodes['Capture Attribute'].outputs[1], tree.nodes['Sample Index'].inputs[3])
                    tree.links.new(tree.nodes['Delete Geometry'].outputs[0], tree.nodes['Store Named Attribute'].inputs[0])
                    tree.links.new(tree.nodes['Capture Attribute'].outputs[2], tree.nodes['Sample Index'].inputs[1])
                    tree.links.new(tree.nodes['Named Attribute.002'].outputs[1], tree.nodes['Capture Attribute'].inputs[2])
                    tree.links.new(tree.nodes['Sample Index'].outputs[0], tree.nodes['Store Named Attribute'].inputs[4])
                    tree.links.new(tree.nodes['Store Named Attribute'].outputs[0], tree.nodes['Group Output'].inputs[0])
                    tree.links.new(tree.nodes['Group Input.001'].outputs[1], tree.nodes['Collection Info'].inputs[0])

            return bpy.data.node_groups['|DJH_⛰_BakePrep']
        
        def GetWaterNode():
            if not '|DJH_⛰_GetWater' in bpy.data.node_groups:
                group = bpy.data.node_groups.new('|DJH_⛰_GetWater', 'GeometryNodeTree')
                group.is_modifier = True


                if bpy.app.version[0] == 4 and bpy.app.version[1] > 0:
                    group.interface.new_socket(socket_type='NodeSocketGeometry', name='Geometry', in_out='OUTPUT')
                    group.interface.new_socket(socket_type='NodeSocketGeometry', name='Geometry', in_out='INPUT')

                    n0 = group.nodes.new('NodeGroupInput')
                    n0.name = 'Group Input'
                    n0.location = Vector((225.760498046875,76.55937957763672))

                    n0 = group.nodes.new('NodeGroupOutput')
                    n0.name = 'Group Output'
                    n0.location = Vector((705.8251342773438,63.97911071777344))

                    n0 = group.nodes.new('GeometryNodeSeparateGeometry')
                    n0.name = 'Separate Geometry'
                    n0.location = Vector((431.6419677734375,59.979156494140625))
                    n0.domain = 'FACE'

                    n0 = group.nodes.new('FunctionNodeCompare')
                    n0.name = 'Equal'
                    n0.location = Vector((233.52890014648438,-14.206085205078125))
                    n0.data_type = 'INT'
                    n0.operation = 'EQUAL'
                    n0.mode = 'ELEMENT'
                    n0.inputs[3].default_value = 1

                    n0 = group.nodes.new('GeometryNodeInputNamedAttribute')
                    n0.name = 'Named Attribute'
                    n0.location = Vector((27.945632934570312,-54.942874908447266))
                    n0.data_type = 'INT'
                    n0.inputs[0].default_value = 'terrain_type'


                    group.links.new(group.nodes['Group Input'].outputs[0], group.nodes['Separate Geometry'].inputs[0])
                    group.links.new(group.nodes['Separate Geometry'].outputs[0], group.nodes['Group Output'].inputs[0])
                    group.links.new(group.nodes['Equal'].outputs[0], group.nodes['Separate Geometry'].inputs[1])
                    group.links.new(group.nodes['Named Attribute'].outputs[0], group.nodes['Equal'].inputs[2])


                else:
                    group.interface.new_socket(socket_type='NodeSocketGeometry', name='Geometry', in_out='OUTPUT')
                    group.interface.new_socket(socket_type='NodeSocketGeometry', name='Geometry', in_out='INPUT')


                    n = group.nodes.new('NodeGroupInput')
                    n.name = 'Group Input'
                    n.location = Vector((225.760498046875,76.55937957763672))

                    n = group.nodes.new('NodeGroupOutput')
                    n.name = 'Group Output'
                    n.location = Vector((987.7073974609375,50.81420135498047))

                    n = group.nodes.new('GeometryNodeSeparateGeometry')
                    n.name = 'Separate Geometry'
                    n.location = Vector((431.6419677734375,56.38872528076172))
                    n.domain = 'FACE'

                    n = group.nodes.new('FunctionNodeCompare')
                    n.name = 'Equal'
                    n.location = Vector((-4.1599731445312, 2.54925537109375))
                    n.data_type = 'INT'
                    n.operation = 'EQUAL'
                    n.inputs[3].default_value = 1

                    n = group.nodes.new('GeometryNodeInputNamedAttribute')
                    n.name = 'Named Attribute'
                    n.location = Vector((-206.1599731445312,-52.54925537109375))
                    n.data_type = 'INT'
                    n.inputs[0].default_value = 'terrain_type'
                    
                    group.links.new(group.nodes['Group Input'].outputs[0], group.nodes['Separate Geometry'].inputs[0])
                    group.links.new(group.nodes['Separate Geometry'].outputs[0], group.nodes['Group Output'].inputs[0])
                    group.links.new(group.nodes['Equal'].outputs[0], group.nodes['Separate Geometry'].inputs[1])
                    group.links.new(group.nodes['Named Attribute'].outputs[4], group.nodes['Equal'].inputs[2])

            return bpy.data.node_groups['|DJH_⛰_GetWater']
        
        def get_trimmed_attrib(name, size, count):    
            # print(f'-getting {name} attribute from mesh')

            val = 0.0
            if name == 'half': val = 0.5
            if name == 'one': val = 1.0

            attrib = [val] * count    

            if name in eval.data.attributes:
                if name not in ['none', 'half', 'one']:
                    eval.data.attributes[name].data.foreach_get("value", attrib)
                
                if name in ['curvature', 'heading']:                
                    attrib = [i * 0.5 + 0.5 for i in attrib]    

            #trim
            # del attrib[size::size+1]
            # del attrib[size*size:]
                        
            return attrib
        
        def set_pixels_and_save(obj, img, pixels, name, ext):
            img.pixels.foreach_set(pixels)
                
            print(f' saving: {obj.name}_{name}.{ext}\n')
            if bpy.context.scene.djh_terranode_props.resolution != 0:
                img.scale(bpy.context.scene.djh_terranode_props.resolution, bpy.context.scene.djh_terranode_props.resolution)
            img.filepath_raw = bpy.path.abspath(os.path.join(bpy.context.scene.djh_terranode_props.filepath, f'{obj.name}_{name}.{ext}'))
            img.save()
        

        #create export collection
        export_col = bpy.data.collections.new("DJH_FBXExport")
        context.scene.collection.children.link(export_col)
        context.view_layer.active_layer_collection = bpy.context.view_layer.layer_collection.children[export_col.name]

        collection = bpy.context.object.users_collection[0]
        dep_graph = bpy.context.evaluated_depsgraph_get()

        names = []

        for obj in collection.objects:
            print(f'{obj.name}:  START')
            base = None
            for m in obj.modifiers:
                if m.type == 'NODES':
                    if m.node_group.name == '  ⛰ Base':
                        base = m
                        break
            
            if base == None:
                print('--not a terrain, skipping')
                continue        
            
            names.append(obj.name)

            print('-copy result to temp')
            temp_mesh = bpy.data.meshes.new_from_object(obj.evaluated_get(dep_graph), preserve_all_data_layers=True, depsgraph=bpy.context.evaluated_depsgraph_get())
            temp_obj = bpy.data.objects.new('DJH_terranode_temp', temp_mesh)
            temp_obj.location = obj.location
            export_col.objects.link(temp_obj)
            # bpy.context.scene.collection.objects.link(temp_obj) # need to link, in order to evaluate??            
            
            print('-remove terrain features')
            modifier = temp_obj.modifiers.new('⛰_BakePrep', "NODES")
            modifier.node_group = GetCleanupNode()
            modifier['Socket_2'] = collection

            print('-evaluate')
            eval = temp_obj.evaluated_get(bpy.context.evaluated_depsgraph_get())

            count = len(eval.data.vertices)    
            size = int(sqrt(count))
                        
            def reset_img():
                if 'terranode_bake' in bpy.data.images:
                    print('--deleting existing image')
                    bpy.data.images.remove( bpy.data.images['terranode_bake'] )
                    
                return bpy.data.images.new('terranode_bake', size, size, alpha=True, float_buffer=True)  

            print(count, size)         

            if bpy.context.preferences.addons[__package__].preferences.use_raw:
                img = reset_img()

                pixels = [1.] * size * size * 4
                pixels[0::4] = get_trimmed_attrib('elevation', size, count)
                img.pixels.foreach_set(pixels)
                
                print(int(bpy.context.preferences.addons[__package__].preferences.raw_size))
                img.scale(int(bpy.context.preferences.addons[__package__].preferences.raw_size)-1, int(bpy.context.preferences.addons[__package__].preferences.raw_size)-1)
                
                print(img.name)
                print(img.size[0], img.size[1])
                pixels = [1.] * len(img.pixels)
                img.pixels.foreach_get(pixels)
                
                # dumb, but blender doesn't allow us to create a 1 channel image...
                del pixels[1::4]
                del pixels[1::3]
                del pixels[1::2]
                                
                ext = 'r16' if bpy.context.preferences.addons[__package__].preferences.export_target == 'unreal' else 'raw'
                
                dir = bpy.path.abspath(bpy.context.scene.djh_terranode_props.filepath)
                if not os.path.exists(dir):
                    os.makedirs(dir)

                file = open(os.path.join(dir, f'{obj.name}.{ext}'), 'wb')
                for p in pixels:
                    file.write(struct.pack('<H', int(p * 32767)))
                    #65535
                file.close()
                
            for output in context.scene.djh_terranode_props.outputs:
                img = reset_img()

                print(f'-rendering image {output.name}:')                
                pixels = [1.] * size * size * 4

                print(f'  R -> {output.red}')
                attrib = get_trimmed_attrib(output.red, size, count)
                pixels[0::4] = attrib    

                print(f'  G -> {output.green}')
                attrib = get_trimmed_attrib(output.green, size, count)
                pixels[1::4] = attrib    

                print(f'  B -> {output.blue}')
                attrib = get_trimmed_attrib(output.blue, size, count)
                pixels[2::4] = attrib    

                print(f'  A -> {output.alpha}')
                attrib = get_trimmed_attrib(output.alpha, size, count)
                pixels[3::4] = attrib    
            
                set_pixels_and_save(obj, img, pixels, output.name, bpy.context.preferences.addons[__package__].preferences.file_format)

            if bpy.context.scene.djh_terranode_props.export_water:
                modifier.node_group = GetWaterNode()
                print('-evaluate water')
                temp_obj.name = f"{obj.name}_water"
                
                bpy.ops.export_scene.fbx(
                    filepath= os.path.join(bpy.path.abspath(bpy.context.scene.djh_terranode_props.filepath), f"{obj.name}_water.fbx"), 
                    check_existing=True, 
                    filter_glob="*.fbx", 
                    use_selection=False, 
                    use_active_collection=True, 
                    global_scale=1, 
                    apply_unit_scale=True, 
                    apply_scale_options='FBX_SCALE_ALL', 
                    use_space_transform=False, 
                    bake_space_transform=True, 
                    object_types={'EMPTY', 'CAMERA', 'LIGHT', 'ARMATURE', 'MESH', 'OTHER'}, 
                    use_mesh_modifiers=True, 
                    use_mesh_modifiers_render=True, 
                    mesh_smooth_type='OFF', 
                    use_subsurf=False, 
                    use_mesh_edges=False, 
                    use_tspace=False, 
                    use_custom_props=False, 
                    add_leaf_bones=False, 
                    primary_bone_axis='Y', 
                    secondary_bone_axis='X', 
                    use_armature_deform_only=False, 
                    armature_nodetype='NULL', 
                    bake_anim=False, 
                    bake_anim_use_all_bones=True, 
                    bake_anim_use_nla_strips=True, 
                    bake_anim_use_all_actions=True, 
                    bake_anim_force_startend_keying=True, 
                    bake_anim_step=1, 
                    bake_anim_simplify_factor=1, 
                    path_mode='AUTO', 
                    embed_textures=False, 
                    batch_mode='OFF', 
                    use_batch_own_dir=True, 
                    use_metadata=True, 
                    axis_forward='-Y', 
                    axis_up='Z')

            print('-cleanup temp mesh')
            bpy.data.meshes.remove(temp_mesh)

        if bpy.context.preferences.addons[__package__].preferences.use_tarr:
            names.sort()
            for output in context.scene.djh_terranode_props.outputs:
                dir = bpy.path.abspath(bpy.context.scene.djh_terranode_props.filepath)
                if not os.path.exists(dir):
                    os.makedirs(dir)

                file = open(os.path.join(dir, f'{collection.name}_{output.name}.tarr'), 'wt')
                for name in names:
                    file.write(f"{name}_{output.name}\n")
                file.close()
            
        #delete the export collection
        bpy.data.collections.remove(export_col)

        return {'FINISHED'}

classes.append(DJH_OT_ExportTerraNode)


class DJH_OT_TerraNodeAddFile(bpy.types.Operator):
    """Adds and item to the collection."""
    bl_idname = 'djh.terranode_add_file'
    bl_label = 'New Export Layer'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        context.scene.djh_terranode_props.outputs.add()        
        context.scene.djh_terranode_props.outputs_index = len(context.scene.djh_terranode_props.outputs) - 1
        return {'FINISHED'}

classes.append(DJH_OT_TerraNodeAddFile)

class DJH_OT_TerraNodeRemoveFile(bpy.types.Operator):
    """Removes an item from the collection."""
    bl_idname = 'djh.terranode_remove_file'
    bl_label = 'Remove Export Layer'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        context.scene.djh_terranode_props.outputs.remove(context.scene.djh_terranode_props.outputs_index)        
        if context.scene.djh_terranode_props.outputs_index > 0:
            context.scene.djh_terranode_props.outputs_index -= 1    
        return {'FINISHED'}

classes.append(DJH_OT_TerraNodeRemoveFile)

class DJH_OT_TerraNodeReadAttributes(bpy.types.Operator):
    """Store which terrain attributes can be saved from the terrain."""
    bl_idname = 'djh.terranode_get_attributes'
    bl_label = 'Get Terrain Attributes'
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        items = context.scene.djh_terranode_props.attributes
        items.clear()

        obj = context.object.evaluated_get(context.evaluated_depsgraph_get())

        for a in obj.data.attributes:
            if a.name[0] == '.': continue
            if a.domain != 'POINT': continue
            if a.data_type not in ['FLOAT', 'INT', 'BOOLEAN']: continue
            
            items.add().set(a.name, a.name, "A terrain attribute")

        if len(context.scene.djh_terranode_props.outputs) == 0:
            check_default_values()

        return {'FINISHED'}

classes.append(DJH_OT_TerraNodeReadAttributes)

class DJH_MT_MenuTerraNode(bpy.types.Menu):
    bl_idname = 'DJH_MT_terranode'
    bl_label = 'TerraNode'

    def draw(self, context):
        l = self.layout
        
        l.operator('djh.terranode_export', text='Export TerraNode')

classes.append(DJH_MT_MenuTerraNode)


class DJH_PT_TerraNode(bpy.types.Panel):
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_idname = "POSE_PT_terranode"
    bl_label = "DJH TerraNode"
    bl_category = 'DJH'

    def draw(self, context):
        layout = self.layout

        not_a_terrain = True
        if context.object:            
            if context.object.modifiers:
                for m in context.object.modifiers:
                    if m.type == 'NODES' and m.node_group.name == '  ⛰ Base':
                        not_a_terrain = False

        if not_a_terrain:
            row = layout.row()
            row.label(text='在偏好设置中点击-添加本插件资产库')
            row = layout.row()
            row.label(text='select a terrain object')
            return
        
        row = layout.row()
        row.operator('djh.terranode_auto_name', text='Auto-Name Terrains')

        row = layout.row()
        row.label(text=f'{context.object.users_collection[0].name} -> {context.object.name}: ({len(context.scene.djh_terranode_props.attributes)} attributes)')
        row = layout.row()
        row.operator('djh.terranode_get_attributes', text='Read Attributes')

        if len(bpy.context.scene.djh_terranode_props.attributes) == 0:
            return
        
        layout.separator()
        row = layout.row()
        row.label(text='Outputs:')

        row = layout.row()
        #  context.preferences.addons[__name__].preferences
        row.template_list('DJH_UL_files', 'djh_tn_export_formats', context.scene.djh_terranode_props, 'outputs', context.scene.djh_terranode_props, 'outputs_index')    

        row = layout.row()
        row.operator('djh.terranode_add_file', text="+")
        row.operator('djh.terranode_remove_file', text="-")

        if len(context.scene.djh_terranode_props.outputs) > 0:
            active_output = context.scene.djh_terranode_props.outputs[context.scene.djh_terranode_props.outputs_index]
            if active_output:
                row = layout.row()        
                row.prop(active_output, 'name')
                row = layout.row(align=True, heading="RGBA")
                row.prop(active_output, 'red', text="")
                row.prop(active_output, 'green', text="")
                row.prop(active_output, 'blue', text="")
                row.prop(active_output, 'alpha', text="")

        row = layout.row()    
        row.prop(context.scene.djh_terranode_props, 'resolution')
        row = layout.row()    
        row.prop(context.scene.djh_terranode_props, 'export_water')
        row = layout.row()    
        row.label(text="Export Path:")
        row = layout.row()
        row.prop(context.scene.djh_terranode_props, 'filepath', icon='FILEBROWSER', text="")
        row = layout.row()
        row.operator('djh.terranode_export', text="Export Terrain")

classes.append(DJH_PT_TerraNode)

# A custom property group for properties related to saving an image
class DJH_PROP_TerraNode_FileFormat(bpy.types.PropertyGroup):
    
    def get_items(self, context):
        items = []
        for a in bpy.context.scene.djh_terranode_props.attributes:
            items.append(a.get())
        items.append(('none', '<NONE>', "a constant value"))
        items.append(('half', '<HALF>', "a constant value"))
        items.append(('one', '<ONE>', "a constant value"))
        return items

    name: StringProperty( name="Suffix", description="Describes the data in the image to save.", default="heightmap" )
    red: EnumProperty(
            name="Red", 
            items=get_items,
            description="Attribute to be used when exporting the red channel.", 
            default=0
        )
    green: EnumProperty(
            name="Green", 
            items=get_items,
            description="Attribute to be used when exporting the green channel.", 
            default=0
        )
    blue: EnumProperty(
            name="Blue", 
            items=get_items,
            description="Attribute to be used when exporting the blue channel.", 
            default=0
        )
    alpha: EnumProperty(
            name="Alpha", 
            items=get_items,
            description="Attribute to be used when exporting the alpha channel.", 
            default=0
        )
classes.append(DJH_PROP_TerraNode_FileFormat)


class DJH_PROP_String(bpy.types.PropertyGroup):
    value: StringProperty(default="elevation")
classes.append(DJH_PROP_String)

class DJH_PROP_Tuple(bpy.types.PropertyGroup):
    a: StringProperty(default="none")
    b: StringProperty(default="<NONE>")
    c: StringProperty(default="A terrain attribute")

    def set(self, a, b, c):
        self.a = a
        self.b = b
        self.c = c

    def get(self):
        return (self.a, self.b, self.c)
    
classes.append(DJH_PROP_Tuple)

class DJH_PROP_TerraNode_PerScene(bpy.types.PropertyGroup):
    filepath: StringProperty(default="default_value")
    outputs: CollectionProperty(type=DJH_PROP_TerraNode_FileFormat, name='Channels')
    outputs_index: IntProperty(name = "Index", default = 0)
    attributes: CollectionProperty(type=DJH_PROP_Tuple, name='Attributes') 
    resolution: IntProperty(name = "Resolution", default = 256)
    export_water: BoolProperty(name="Export Water Meshes", default=False)
    # [('none', '<NONE>', "A terrain attribute")]
classes.append(DJH_PROP_TerraNode_PerScene)

class DJH_UL_files(bpy.types.UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname):
        if item.red:
            layout.prop(item, "name", text="", emboss=False, icon_value=icon)
        else:
            layout.label(text="", translate=False, icon_value=icon)
classes.append(DJH_UL_files)

class DJH_UL_Value(bpy.types.UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname):
        if item.value:
            layout.prop(item, "value", text="", emboss=False, icon_value=icon)
classes.append(DJH_UL_Value)

class DJH_TerraNode_Prefs(bpy.types.AddonPreferences):
    bl_idname = __package__


    export_target: EnumProperty(
        name="Export Target", 
        default='none',
        description="Optimize the export for another software.",
        items={
            ('none',"Generic",""),
            ('unity',"Unity",""),
            ('unreal',"Unreal","")
        })
    
    file_format: EnumProperty(
        name="Export Format", 
        default='png',
        description="The format to use when saving terrain tiles",
        items={
            ('png',"PNG",""),
            ('exr',"EXR",""),
            ('tga',"TGA",""),
            ('tif',"TIF","")
        })
    
    use_raw: BoolProperty(
        name="Export .RAW Heightmap", 
        default=False,
        description="Game engines often import heightmaps in a .raw format, in specific sizes, with a single channel."
    )
    
    raw_size: EnumProperty(
        name="RAW Size", 
        default='256',
        description="The resolution to scale the RAW heightmap to",
        items={
            ('64',"64",""),
            ('128',"128",""),
            ('256',"256",""),
            ('512',"512",""),
            ('1024',"1024","")
        })
    
    use_tarr: BoolProperty(
        name="Export .tarr Texture Array Lists", 
        default=False,
        description="Unity shaders can use texture arrays but they must be created via a script.  When checked, blender will save lists of texture names to .tarr files which can be converted to texture arrays on import"
    )

    def draw(self, context):
        layout = self.layout
        row = layout.row()
        row.prop(self, 'export_target')
        row = layout.row()
        row.prop(self, 'file_format')
        row = layout.row()
        row.prop(self, 'use_raw')
        row = layout.row()
        row.prop(self, 'raw_size')
        row = layout.row()
        row.prop(self, 'use_tarr')
        row = layout.row()
        row.operator("terranode.add_lib_assets",text="添加TerraNode到资产库")

classes.append(DJH_TerraNode_Prefs)

    
class Add_default_lib_Asssets(bpy.types.Operator):
    bl_idname = "terranode.add_lib_assets"
    bl_label = "Add Default Library Assets"
    bl_description = ""
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        return True

    def execute(self, context):

        lib_name = "TerraNode_library"
        lib_path = str(Path(__file__).parent/lib_name)
        app_lib_data = {}
        for lib in bpy.context.preferences.filepaths.asset_libraries:
            app_lib_data[lib.name] = lib.path

        asset_libraries = bpy.context.preferences.filepaths.asset_libraries
        if lib_name not in app_lib_data:
            bpy.ops.preferences.asset_library_add(directory=lib_path)
            asset_libraries[-1].name = lib_name
            self.report({"INFO"}, "TerraNode资产库已添加")

        else:
            self.report({"INFO"}, "TerraNode资产库已存在，已更新路径")
            asset_libraries.get(lib_name).path = lib_path
        return {"FINISHED"}
classes.append(Add_default_lib_Asssets)

def check_default_values():        
    if bpy.context.scene.djh_terranode_props.filepath == 'default_value':
        bpy.context.scene.djh_terranode_props.filepath = os.path.join("//terranode", "")

    if len(bpy.context.scene.djh_terranode_props.attributes) == 0:
        return

    outputs = bpy.context.scene.djh_terranode_props.outputs

    if len(outputs) == 0:
        o = outputs.add()
        o.name = 'heightmap'
        o.alpha = 'one'
        o = outputs.add()
        o.name = 'attributes' 
        o.red = 'steepness' if 'steepness' in o.red else 'one'
        o.green = 'curvature' if 'curvature' in o.green else 'one'
        o.blue = 'erosion' if 'erosion' in o.blue else 'one'
        o.alpha = 'one'
        o = outputs.add()
        o.name = 'features'
        o.red = 'roads' if 'roads' in o.red else 'one'
        o.green = 'water_level' if 'water_level' in o.green else 'one'
        o.blue = 'none'
        o.alpha = 'one'

def onRegister():
    check_default_values()
    return None

@persistent
def onLoad(scene):
    check_default_values()

def register():
    def draw_menu_terranode(self, context):
        layout = self.layout
        layout.menu("DJH_MT_terranode")

    for c in classes:
        bpy.utils.register_class(c)

    bpy.types.VIEW3D_MT_object.append(draw_menu_terranode)
    Scene.djh_terranode_props = PointerProperty(type=DJH_PROP_TerraNode_PerScene)
    
    bpy.app.timers.register(onRegister, first_interval=0)
    bpy.app.handlers.load_post.append(onLoad)

def unregister():
    for c in classes:
        bpy.utils.unregister_class(c)

    del Scene.djh_terranode_props
    bpy.app.handlers.load_post.remove(onLoad)

if __name__ == '__main__':
    register()
